@@ -1,13 +1,27 @@
+# ----------------------------------------------------------------------
+# NAME       : BibTeX/Entry.pm
+# CLASSES    : Text::BibTeX::Entry
+# RELATIONS  : base class for Text::BibTeX::StructuredEntry, and 
+#              ultimately for all user-supplied structured entry classes
+# DESCRIPTION: Provides an object-oriented interface to BibTeX entries.
+# CREATED    : March 1997, Greg Ward
+# MODIFIED   : 
+# VERSION    : $Id: Entry.pm,v 1.17 1999/03/11 04:54:06 greg Exp $
+# COPYRIGHT  : Copyright (c) 1997-98 by Gregory P. Ward.  All rights
+#              reserved.
+# 
+#              This file is part of the Text::BibTeX library.  This
+#              library is free software; you may redistribute it and/or
+#              modify it under the same terms as Perl itself.
+# ----------------------------------------------------------------------
 package Text::BibTeX::Entry;
 
 require 5.004;                          # for isa, and delete on a slice
 
-# $Id: Entry.pm,v 1.9 1997/10/05 23:46:28 greg Exp $
-
 use strict;
-use Carp;
 use UNIVERSAL 'isa';
-import Text::BibTeX qw(:metatypes);
+use Carp;
+import Text::BibTeX qw(:metatypes :nodetypes);
 
 =head1 NAME
 
@@ -15,6 +29,8 @@ Text::BibTeX::Entry - read and parse BibTeX files
 
 =head1 SYNOPSIS
 
+   use Text::BibTeX;            # do not use Text::BibTeX::Entry alone!
+
    # ...assuming that $bibfile and $newbib are both objects of class
    # Text::BibTeX::File, opened for reading and writing (respectively):
 
@@ -35,7 +51,7 @@ Text::BibTeX::Entry - read and parse BibTeX files
    $type = $entry->type;
 
    # if metatype is BTE_REGULAR or BTE_MACRODEF:
-   $key = $entry->key;                  # BTE_REGULAR only, actually
+   $key = $entry->key;                  # only for BTE_REGULAR metatype
    $num_fields = $entry->num_fields;
    @fieldlist = $entry->fieldlist;
    $has_title = $entry->exists ('title');
@@ -71,17 +87,23 @@ Text::BibTeX::Entry - read and parse BibTeX files
 
 =head1 DESCRIPTION
 
-C<Text::BibTeX::Entry> does all the real work of reading and parsing
+F<Text::BibTeX::Entry> does all the real work of reading and parsing
 BibTeX files.  (Well, actually it just provides an object-oriented Perl
 front-end to a C library that does all that.  But that's not important
 right now.)
 
-BibTeX entries can be read either from C<Text::BibTeX::File> objects (using
+BibTeX entries can be read either from F<Text::BibTeX::File> objects (using
 the C<read> method), or directly from a filehandle (using the C<parse>
 method), or from a string (using C<parse_s>).  The first is preferable,
-since you don't have to worry about supplying the filename, and because I
-might add more functionality to that method in the future.  Currently,
-though, the two are pretty much identical.
+since you don't have to worry about supplying the filename, and because of
+the extra functionality provided by the F<Text::BibTeX::File> class.
+Currently, this means that you may specify the I<database structure> to
+which entries are expected to conform via the F<File> class.  This lets you
+ensure that entries follow the rules for required fields and mutually
+constrained fields for a particular type of database, and also gives you
+access to all the methods of the I<structured entry class> for this
+database structure.  See L<Text::BibTeX::Structure> for details on database
+structures.
 
 Once you have the entry, you can query it or change it in a variety of
 ways.  The query methods are C<parse_ok>, C<type>, C<key>, C<num_fields>,
@@ -89,12 +111,12 @@ C<fieldlist>, C<exists>, and C<get>.  Methods for changing the entry are
 C<set_type>, C<set_key>, C<set_fieldlist>, C<delete>, and C<set>.
 
 Finally, you can output BibTeX entries, again either to an open
-C<Text::BibTeX::File> object, a filehandle or a string.  (A
-C<Text::BibTeX::File> object or filehandle must, of course, have been
-opened in write mode.)  Output to a C<Text::BibTeX::File> object is done
-with the C<write> method, to a filehandle via C<print>, and to a string
-with C<print_s>.  Again, the nice object-oriented way of doing things is
-recommended for future extensibility.
+F<Text::BibTeX::File> object, a filehandle or a string.  (A filehandle or
+F<File> object must, of course, have been opened in write mode.)  Output to
+a F<File> object is done with the C<write> method, to a filehandle via
+C<print>, and to a string with C<print_s>.  Using the F<File> class is
+recommended for future extensibility, although it currently doesn't offer
+anything extra.
 
 =head1 METHODS
 
@@ -104,16 +126,16 @@ recommended for future extensibility.
 
 =item new ([SOURCE])
 
-Creates a new C<Text::BibTeX::Entry> object.  If the SOURCE parameter is
-supplied, it must be one of the following: a C<Text::BibTeX::File> (or
+Creates a new F<Text::BibTeX::Entry> object.  If the SOURCE parameter is
+supplied, it must be one of the following: a F<Text::BibTeX::File> (or
 descendant class) object, a filename/filehandle pair, or a string.  Calls
-C<read> to read from a C<Text::BibTeX::File> object, C<parse> to read from
+C<read> to read from a F<Text::BibTeX::File> object, C<parse> to read from
 a filehandle, and C<parse_s> to read from a string.
 
 A filehandle can be specified as a GLOB reference, or as an
 C<IO::Handle> (or descendants) object, or as a C<FileHandle> (or
 descendants) object.  (But there's really no point in using
-C<FileHandle> objects, since C<Text::BibTeX> requires Perl 5.004, which
+C<FileHandle> objects, since F<Text::BibTeX> requires Perl 5.004, which
 always includes the C<IO> modules.)  You can I<not> pass in the name of
 a filehandle as a string, though, because F<Text::BibTeX::Entry>
 conforms to the C<use strict> pragma (which disallows such symbolic
@@ -136,7 +158,7 @@ C<$filename> (error handling ignored):
    $file = new IO::File $filename;
    $entry = new Text::BibTeX::Entry ($filename, $file);
 
-But using a C<Text::BibTeX::File> object is preferred:
+But using a F<Text::BibTeX::File> object is simpler and preferred:
 
    $file = new Text::BibTeX::File $filename;
    $entry = new Text::BibTeX::Entry $file;
@@ -165,13 +187,13 @@ sub new
 
       if (@source == 1 && isa ($source[0], 'Text::BibTeX::File'))
       { 
-         # XXX err... what if $file doesn't have a structure, or the
-         # structure doesn't have an entry_class???
-
          my $file = $source[0];
          $status = $self->read ($file);
-         bless $self, $file->structure->entry_class
-            if ($file->structure);
+         if (my $structure = $file->structure)
+         {
+            $self->{structure} = $structure;
+            bless $self, $structure->entry_class;
+         }            
       }
       elsif (@source == 2 && ! ref $source[0] && fileno ($source[1]))
           { $status = $self->parse ($source[0], $source[1]) }
@@ -190,7 +212,7 @@ sub new
 =item read (BIBFILE)
 
 Reads and parses an entry from BIBFILE, which must be a
-C<Text::BibTeX::File> object (or descendant).  The next entry will be read
+F<Text::BibTeX::File> object (or descendant).  The next entry will be read
 from the file associated with that object.
 
 Returns the same as C<parse> (or C<parse_s>): false if no entry found
@@ -201,7 +223,7 @@ Returns the same as C<parse> (or C<parse_s>): false if no entry found
 
 sub read
 {
-   my ($self, $source) = @_;
+   my ($self, $source, $preserve) = @_;
    croak "`source' argument must be ref to open Text::BibTeX::File " .
          "(or descendant) object"
       unless (isa ($source, 'Text::BibTeX::File'));
@@ -209,7 +231,7 @@ sub read
    my $fn = $source->{'filename'};
    my $fh = $source->{'handle'};
    $self->{'file'} = $source;        # store File object for later use
-   return $self->parse ($fn, $fh);
+   return $self->parse ($fn, $fh, $preserve);
 }
 
 
@@ -222,7 +244,7 @@ but not the second one -- it's pushed back onto the input stream for the
 next entry] is parsed as a BibTeX entry, with the simultaneous
 construction of an abstract syntax tree [AST].  The AST is traversed to
 ferret out the most interesting information, and this is stuffed into a
-Perl hash, which coincidentally is the C<Text::BibTeX::Entry> object
+Perl hash, which coincidentally is the F<Text::BibTeX::Entry> object
 you've been tossing around.  But you don't need to know any of that -- I
 just figured if you've read this far, you might want to know something
 about the inner workings of this module.)
@@ -258,14 +280,41 @@ putting multiple entries in one string.
 
 =cut
 
-# see BibTeX.xs for the implementation of the `parse' and `parse_s' methods
+sub _preserve
+{
+   my ($self, $preserve) = @_;
+
+   $preserve = $self->{'file'}->preserve_values
+      if ! defined $preserve && 
+         defined $self->{'file'} &&
+         isa ($self->{'file'}, 'Text::BibTeX::File');
+   require Text::BibTeX::Value if $preserve;
+   $preserve;
+}
+
+sub parse
+{
+   my ($self, $filename, $filehandle, $preserve) = @_;
+
+   $preserve = $self->_preserve ($preserve);
+   _parse ($self, $filename, $filehandle, $preserve);
+}
+
+
+sub parse_s
+{
+   my ($self, $text, $preserve) = @_;
+
+   $preserve = $self->_preserve ($preserve);
+   _parse_s ($self, $text, $preserve);
+}
 
 
 =head2 Entry query methods
 
 =over 4
 
-=item parse_ok
+=item parse_ok ()
 
 Returns false if there were any serious errors encountered while parsing
 the entry.  (A "serious" error is a lexical or syntax error; currently,
@@ -273,35 +322,35 @@ warnings such as "undefined macro" result in an error message being
 printed to C<stderr> for the user's edification, but no notice is
 available to the calling code.)
 
-=item type
+=item type ()
 
 Returns the type of the entry.  (The `type' is the word that follows the
 '@' sign; e.g. `article', `book', `inproceedings', etc. for the standard
 BibTeX styles.)
 
-=item metatype
+=item metatype ()
 
 Returns the metatype of the entry.  (The `metatype' is a numeric value used
 to classify entry types into four groups: comment, preamble, macro
 definition (C<@string> entries), and regular (all other entry types).
-Text::BibTeX exports four constants for these metatypes: BTE_COMMENT,
-BTE_PREAMBLE, BTE_MACRODEF, and BTE_REGULAR.)
+F<Text::BibTeX> exports four constants for these metatypes: C<BTE_COMMENT>,
+C<BTE_PREAMBLE>, C<BTE_MACRODEF>, and C<BTE_REGULAR>.)
 
-=item key
+=item key ()
 
 Returns the key of the entry.  (The key is the token immediately
 following the opening `{' or `(' in "regular" entries.  Returns C<undef>
 for entries that don't have a key, such as macro definition (C<@string>)
 entries.)
 
-=item num_fields
+=item num_fields ()
 
 Returns the number of fields in the entry.  (Note that, currently, this is
 I<not> equivalent to putting C<scalar> in front of a call to C<fieldlist>.
 See below for the consequences of calling C<fieldlist> in a scalar
 context.)
 
-=item fieldlist
+=item fieldlist ()
 
 Returns the list of fields in the entry.  In a scalar context, returns a
 reference to the object's own list of fields.  That way, you can change or
@@ -344,11 +393,13 @@ syntax errors in the input, or if you pass an undefined value to C<set>, or
 if you create a new field with C<set_fieldlist> (the new field's value is
 implicitly set to C<undef>).
 
-Currently, the field value is what the input looks like after "maximal
+Normally, the field value is what the input looks like after "maximal
 processing"--quote characters are removed, whitespace is collapsed (the
 same way that BibTeX itself does it), macros are expanded, and multiple
-tokens are pasted together.  For example, if your input file has the
-following:
+tokens are pasted together.  (See L<bt_postprocess> for details on the
+post-processing performed by B<btparse>.)
+
+For example, if your input file has the following:
 
    @string{of = "of"}
    @string{foobars = "Foobars"}
@@ -357,16 +408,20 @@ following:
      title = {   The Mating Habits      } # of # " Adult   " # foobars
    }
 
-then querying the value of the C<title> field from the C<foobar> entry
-would give the string "The Mating Habits of Adult Foobars".  I have plans
-up my sleeve for giving access to the data at various stages of processing
-(the underlying C library is quite flexible in this regard; I just have to
-translate the flexibility to Perl), but haven't finalized anything yet.  If
-you have ideas, feel free to email me!  (I also have plans for
-documenting what exactly is done to strings in my BibTeX parser; that'll
-probably be distributed with the C library, btparse.)
+then using C<get> to query the value of the C<title> field from the
+C<foobar> entry would give the string "The Mating Habits of Adult Foobars".
 
-=item value
+However, in certain circumstances you may wish to preserve the values as
+they appear in the input.  This is done by setting a C<preserve_values>
+flag at some point; then, C<get> will return not strings but
+F<Text::BibTeX::Value> objects.  Each F<Value> object is a list of
+F<Text::BibTeX::SimpleValue> objects, which in turn consists of a simple
+value type (string, macro, or number) and the text of the simple value.
+Various ways to set the C<preserve_values> flag and the interface to
+both F<Value> and F<SimpleValue> objects are described in
+L<Text::BibTeX::Value>.
+
+=item value ()
 
 Retuns the single string associated with C<@comment> and C<@preamble>
 entries.  For instance, the entry
@@ -377,6 +432,10 @@ entries.  For instance, the entry
 would return a value of "This is a preamble---the concatenation of
 several strings".
 
+If this entry was parsed in "value preservation" mode, then C<value>
+acts like C<get>, and returns a F<Value> object rather than a simple
+string.
+
 =back
 
 =cut
@@ -417,8 +476,8 @@ Incidentally, both of these methods assume that the strings being split
 have already been "collapsed" in the BibTeX way, i.e. all leading and
 trailing whitespace removed and internal whitespace reduced to single
 spaces.  This should always be the case when using these two methods on
-a C<Text::BibTeX::Entry> object, but these are actually just front ends
-to more general functions in C<Text::BibTeX>.  (More general in that you
+a F<Text::BibTeX::Entry> object, but these are actually just front ends
+to more general functions in F<Text::BibTeX>.  (More general in that you
 supply the string to be parsed, rather than supplying the name of an
 entry field.)  Should you ever use those more general functions
 directly, you might have to worry about collapsing whitespace; see
@@ -433,12 +492,12 @@ have problems with it, think it's just perfect, or whatever.
 
 =item split (FIELD [, DELIM [, DESC]])
 
-Splits the value of FIELD on DELIM (default: `and').  Don't assume that,
-just because the names are the same, this works the same as Perl's
-builtin C<split>: in particular, DELIM must be a simple string (no
-regexps), and delimiters that are at the beginning or end of the string,
-or at non-zero brace depth, or not surrounded by whitespace, are
-ignored.  Some examples might illuminate matters:
+Splits the value of FIELD on DELIM (default: `and').  Don't assume that
+this works the same as Perl's builtin C<split> just because the names are
+the same: in particular, DELIM must be a simple string (no regexps), and
+delimiters that are at the beginning or end of the string, or at non-zero
+brace depth, or not surrounded by whitespace, are ignored.  Some examples
+might illuminate matters:
 
    if field F is...                then split (F) returns...
    'Name1 and Name2'               ('Name1', 'Name2')
@@ -458,14 +517,14 @@ It is only used for generating warning messages.
 =item names (FIELD)
 
 Splits FIELD as described above, and further splits each name into four
-components: first, von, last, and jr.  The rules for this are described
-colloquially in any BibTeX documentation, and will eventually be spelled
-out more formally in the documentation for the F<btparse> library.
+components: first, von, last, and jr.  
 
-Returns a list of structures representing the names.  Each structure is
-a hash with at most four keys (C<first>, C<von>, C<last>, and C<jr>);
-the values are either C<undef> or lists of the tokens that make up that
-component of the name.  For example, the following entry:
+Returns a list of F<Text::BibTeX::Name> objects, each of which represents
+one name.  Use the C<part> method to query these objects; see
+L<Text::BibTeX::Name> for details on the interface to name objects (and on
+name-parsing as well).
+
+For example if this entry:
 
    @article{foo,
             author = {John Smith and 
@@ -473,30 +532,17 @@ component of the name.  For example, the following entry:
                       Ludwig van Beethoven and
                       {Foo, Bar and Company}}}
 
-would result in the following list of name-structures being returned by
-C<names>:
-
-   ( { first => ['John'],
-       von   => undef,
-       last  => ['Smith'],
-       jr    => undef },
-     { first => ['J.', 'Random'],
-       von   => undef,
-       last  => ['Hacker'],
-       jr    => undef },
-     { first => ['Ludwig'],
-       von   => ['van'],
-       last  => ['Beethoven']
-       jr    => undef },
-     { first => undef,
-       von   => undef,
-       last  => ['{Foo, Bar and Company}'],
-       jr    => undef } )
-
-and some example code might look like
+has been parsed into a F<Text::BibTeX::Entry> object C<$entry>, then
 
    @names = $entry->names ('author');
-   $sort_key = $names[0]->{'last'} . ' ' . $names[0]->{'first'};
+
+will put a list of F<Text::BibTeX::Name> objects in C<@names>.  These can
+be queried individually as described in L<Text::BibTeX::Name>; for instance,
+
+   @last = $names[0]->part ('last');
+
+would put the list of tokens comprising the last name of the first author
+into the C<@last> array: C<('Smith')>.
 
 =cut
 
@@ -511,14 +557,16 @@ sub split
    my $filename = ($self->{'file'} && $self->{'file'}{'filename'});
    my $line = $self->{'lines'}{$field};
 
-   local $^W = 0                        # suppress spurious warning from 
-      unless defined $filename;         # undefined $filename
+#   local $^W = 0                        # suppress spurious warning from 
+#      unless defined $filename;         # undefined $filename
    Text::BibTeX::split_list ($self->{'values'}{$field}, $delim,
                              $filename, $line, $desc);
 }
 
 sub names
 {
+   require Text::BibTeX::Name;
+
    my ($self, $field) = @_;
    my (@names, $i);
 
@@ -526,15 +574,17 @@ sub names
    my $line = $self->{'lines'}{$field};
 
    @names = $self->split ($field);
-   local $^W = 0                        # suppress spurious warning from 
-      unless defined $filename;         # undefined $filename
+#   local $^W = 0                        # suppress spurious warning from 
+#      unless defined $filename;         # undefined $filename
    for $i (0 .. $#names)
    {
-      $names[$i] = Text::BibTeX::split_name ($names[$i], $filename, $line, $i);
+      $names[$i] = new Text::BibTeX::Name ($names[$i], $filename, $line, $i);
    }
    @names;
 }
 
+=back
+
 =head2 Entry modification methods
 
 =over 4
@@ -560,6 +610,10 @@ order (i.e. the input is treated like a list, not a hash).  For example:
    $entry->set ('author', $author);
    $entry->set ('author', $author, 'editor', $editor);
 
+VALUE can be either a simple string or a F<Text::BibTeX::Value> object;
+it doesn't matter if the entry was parsed in "full post-processing" or
+"preserve input values" mode.
+
 =item delete (FIELD)
 
 Deletes field FIELD from an entry.
@@ -660,7 +714,7 @@ sub set_fieldlist
 =item write (BIBFILE)
 
 Prints a BibTeX entry on the filehandle associated with BIBFILE (which
-should be a C<Text::BibTeX::File> object, opened for output).  Currently
+should be a F<Text::BibTeX::File> object, opened for output).  Currently
 the printout is not particularly human-friendly; a highly configurable
 pretty-printer will be developed eventually.
 
@@ -668,7 +722,7 @@ pretty-printer will be developed eventually.
 
 Prints a BibTeX entry on FILEHANDLE.
 
-=item print_s
+=item print_s ()
 
 Prints a BibTeX entry to a string, which is the return value.
 
@@ -686,6 +740,7 @@ sub print
 {
    my ($self, $handle) = @_;
 
+   $handle ||= \*STDOUT;
    print $handle $self->print_s;
 }
 
@@ -694,6 +749,30 @@ sub print_s
    my $self = shift;
    my ($field, $output);
 
+   sub value_to_string
+   {
+      my $value = shift;
+
+      if (! ref $value)                 # just a string
+      {
+         return qq["$value"];
+      }
+      else                              # a Text::BibTeX::Value object
+      {
+         confess "value is a reference, but not to Text::BibTeX::Value object"
+            unless isa ($value, 'Text::BibTeX::Value');
+         my @values = $value->values;
+         @values = map
+         { 
+            $_->type == &BTAST_STRING 
+               ? '"' . $_->text . '"'
+               : $_->text;
+         } @values;
+         return join (' # ', @values);            
+      }
+   }
+
+
    carp "entry type undefined" unless defined $self->{'type'};
 
    # Regular and macro-def entries have to be treated differently when
@@ -716,17 +795,27 @@ sub print_s
    # the entire entry, on one line, right here.
    else                                 # comment or preamble
    {
-      return sprintf ("@%s{%s}\n\n", $self->{'type'}, $self->{'value'});
+      return sprintf ("@%s{%s}\n\n",
+                      $self->{'type'},
+                      value_to_string ($self->{'value'}));
    }
 
    # Here we print out all the fields/values of a regular or macro-def entry
-   foreach $field (@{$self->{'fields'}})
+   my @fields = @{$self->{'fields'}};
+   while ($field = shift @fields)
    {
-      carp "field \"$field\" has undefined value\n"
-         unless defined $self->{'values'}{$field};
-      $output .= sprintf ("  %s = {%s},\n",
-                          $field, 
-                          $self->{'values'}{$field} || '');
+      my $value = $self->{'values'}{$field};
+      if (! defined $value)
+      {
+         carp "field \"$field\" has undefined value\n";
+         $value = '';
+      }
+
+      $output .= "  $field = ";
+      $output .= value_to_string ($value);
+
+      $output .= ',' if @fields;        # more fields yet to come
+      $output .= "\n";
    }
 
    # Tack on the last line, and we're done!
@@ -734,6 +823,7 @@ sub print_s
    $output;
 }
 
+=back
 
 =head2 Miscellaneous methods
 
@@ -799,16 +889,67 @@ sub warn
 }
 
 
+=item line ([FIELD])
+
+Returns the line number of FIELD.  If the entry was parsed from a string,
+this still works--it's just the line number relative to the start of the
+string.  If the entry was parsed from a file, this works just as you'd
+expect it to: it returns the absolute line number with respect to the
+whole file.  Line numbers are one-based.
+
+If FIELD is not supplied, returns a two-element list containing the line
+numbers of the beginning and end of the whole entry.  (Actually, the
+"end" line number is currently inaccurate: it's really the the line
+number of the last field in the entry.  But it's better than nothing.)
+
+=cut
+
+sub line
+{
+   my ($self, $field) = @_;
+
+   if (defined $field)
+   {
+      return $self->{'lines'}{$field};
+   }
+   else
+   {
+      return @{$self->{'lines'}}{'START','STOP'};
+   }
+}
+
+=item filename ()
+
+Returns the name of the file from which the entry was parsed.  Only
+works if the file is represented by a F<Text::BibTeX::File> object---if
+you just passed a filename/filehandle pair to C<parse>, you can't get
+the filename back.  (Sorry.)
+
+=cut
+
+sub filename
+{
+   my $self = shift;
+
+   $self->{'file'}{'filename'};         # ooh yuck -- poking into File object 
+}
+
 1;
 
+=back
+
+=head1 SEE ALSO
+
+L<Text::BibTeX>, L<Text::BibTeX::File>, L<Text::BibTeX::Structure>
+
 =head1 AUTHOR
 
-Greg Ward <greg@bic.mni.mcgill.ca>
+Greg Ward <gward@python.net>
 
 =head1 COPYRIGHT
 
-Copyright (c) 1997 by Gregory P. Ward.  All rights reserved.  This is free
-software; you can redistribute it and/or modify it under the same terms as
-Perl itself.
+Copyright (c) 1997-98 by Gregory P. Ward.  All rights reserved.  This file
+is part of the Text::BibTeX library.  This library is free software; you
+may redistribute it and/or modify it under the same terms as Perl itself.
 
 =cut
